import os
import pytest
import relic
import relic.release
import relic.git
import shutil
import stat
import sys
from subprocess import check_call, STDOUT


WIN32 = True if sys.platform.startswith('win') else False


def runner(command):
    with open(os.devnull, 'w') as devnull:
        check_call(command, shell=True, stdout=devnull, stderr=STDOUT)


def touch(filename):
    with open(filename, 'a'):
        pass


onerror = None
if WIN32:
    # From: shutil.rmtree fails on Windows with 'Access is denied'
    # https://stackoverflow.com/q/1213706/1711764
    def onerror(func, path, exc_info):
        """
        Error handler for ``shutil.rmtree``.

        If the error is due to an access error (read only file)
        it attempts to add write permission and then retries.

        If the error is for another reason it re-raises the error.

        Usage : ``shutil.rmtree(path, onerror=onerror)``
        """
        if not os.access(path, os.W_OK):
            # Is the error an access error ?
            os.chmod(path, stat.S_IWUSR)
            func(path)
        else:
            raise


@pytest.fixture()
def _baserepo(tmpdir):
    repo_path = tmpdir.mkdir('repo')
    os.chdir(str(repo_path))

    # Create a basic git repository.
    runner('git init')
    runner('git config user.name nobody')
    runner('git config user.email nobody@nowhere')
    runner('git config commit.gpgsign false')
    touch('testfile')
    runner('git add testfile')
    runner("git commit -m 'Initial'")
    yield


@pytest.mark.usefixtures('_baserepo')
class TestGit(object):
    def test_git_describe_not_a_repository(self):
        shutil.rmtree('.git', onerror=onerror)
        desc = relic.git.git_describe()
        assert desc is None

    def test_git_describe(self):
        runner('git tag -a 1.0.0 -m "test message"')
        desc = relic.git.git_describe()
        assert isinstance(desc, str)

    def test_git_log_date_not_a_repository(self):
        shutil.rmtree('.git', onerror=onerror)
        date = relic.git.git_log_date()
        assert date is None

    def test_git_log_date(self):
        runner('git tag -a 1.0.0 -m "test message"')
        date = relic.git.git_log_date()
        assert isinstance(date, str)
        assert len(date) > 0

    def test_git_version_info_not_a_repository(self):
        shutil.rmtree('.git', onerror=onerror)
        v = relic.git.git_version_info()
        assert v is None

    def test_git_version_info(self):
        runner('git tag -a 1.0.0 -m "test message"')
        v = relic.git.git_version_info()
        assert isinstance(v, relic.git.GitVersion)


@pytest.mark.usefixtures('_baserepo')
class TestRelease(object):
    def test_version_not_a_repository(self):
        shutil.rmtree('.git', onerror=onerror)
        v = relic.release.get_info()
        assert v.pep386 == '0.0.0'
        assert int(v.post) < 0

    def test_version_without_tag(self):
        v = relic.release.get_info()
        assert v is not None
        assert isinstance(v, relic.git.GitVersion)
        assert '0.0.0.dev' in v.pep386
        assert '+' in v.pep386
        assert int(v.post) == 1

    def test_version_without_tag_autocount(self):
        count = 0x10
        for i in range(count):
            filename = 'file{}.txt'.format(i)
            touch(filename)
            runner('git add {}'.format(filename))
            runner('git commit -m "added {}"'.format(filename))

        v = relic.release.get_info()
        assert v is not None
        assert isinstance(v, relic.git.GitVersion)
        assert '0.0.0.dev' in v.pep386
        assert '+' in v.pep386
        assert int(v.post) == count + 1

    def test_version_pep386_release(self):
        runner('git tag -a 1.0.0 -m "test message"')
        v = relic.release.get_info()
        assert v.pep386 == '1.0.0'
        assert not v.dirty
        assert v.post == '0'

    def test_version_pep386_dev(self):
        runner('git tag -a 1.0.0 -m "test message"')
        touch('testfile2')
        runner('git add testfile2')
        runner('git commit -m "add testfile2"')
        v = relic.release.get_info()
        assert '.dev' in v.pep386
        assert not v.dirty
        assert int(v.post) > 0

    def test_version_dirty_true_if_dirty(self):
        touch('testfile3')
        runner('git add testfile3')
        v = relic.release.get_info()
        assert isinstance(v.dirty, bool)
        assert v.dirty

    def test_version_dirty_false_if_not_dirty(self):
        v = relic.release.get_info()
        assert isinstance(v.dirty, bool)
        assert not v.dirty

    def test_version_short_is_short(self):
        runner('git tag -a 1.0.0 -m "test message"')
        touch('testfile3')
        runner('git add testfile3')
        runner('git commit -m "add testfile"')
        v = relic.release.get_info()
        assert isinstance(v.short, str)
        assert '.' in v.short
        assert '-' not in v.short

    def test_version_long_is_long(self):
        runner('git tag -a 1.0.0 -m "test message"')
        touch('testfile3')
        runner('git add testfile3')
        runner('git commit -m "add testfile"')
        v = relic.release.get_info()
        assert isinstance(v.long, str)
        assert '.' in v.long
        assert '-' in v.long

    def test_version_post_incremented_after_commit(self):
        runner('git tag -a 1.0.0 -m "test message"')
        touch('testfile3')
        runner('git add testfile3')
        runner('git commit -m "add testfile"')
        v = relic.release.get_info()
        assert isinstance(v.post, str)
        post = int(v.post)
        assert post > 0

    def test_version_commit_not_a_repository(self):
        shutil.rmtree('.git', onerror=onerror)
        v = relic.release.get_info()
        assert not v.commit

    def test_version_commit(self):
        runner('git tag -a 1.0.0 -m "test message"')
        v = relic.release.get_info()
        assert v.commit

    def test_version_handle_v_prefix(self):
        runner('git tag -a v1.0.0 -m "test message"')
        v = relic.release.get_info()
        assert 'v' not in v.short
        assert v.short == '1.0.0'

    def test_version_handle_purge_release_prefix(self):
        runner('git tag -a release_1.0.0 -m "test message"')
        v = relic.release.get_info()
        assert 'release_' not in v.short
        assert v.short == '1.0.0'

    def test_version_handle_purge_v_and_release_prefix(self):
        runner('git tag -a release_v1.0.0 -m "test message"')
        v = relic.release.get_info()
        assert 'release_' not in v.short
        assert 'v' not in v.short
        assert v.short == '1.0.0'

    def test_version_handle_purge_nightmare_pattern(self):
        runner('git tag -a itsover9000-v1.0.0_codingaward -m "test message"')
        purge_pattern = ['itsover9000-', '_codingaward']
        v = relic.release.get_info(purge_pattern)
        for p in purge_pattern:
            assert p not in v.short

        # Should have been purged via regex
        assert not v.short.startswith('v')
        assert v.short == '1.0.0'

    def test_version_date_not_a_repository(self):
        # Only check if date is populated with "something"
        shutil.rmtree('.git', onerror=onerror)
        v = relic.release.get_info()
        assert isinstance(v.date, str)
        assert not v.date

    def test_version_date(self):
        # Only check if date is populated with "something"
        runner('git tag -a 1.0.0 -m "test message"')
        v = relic.release.get_info()
        assert isinstance(v.date, str)
        assert v.date

    def test_read_relic_info(self):
        runner('git tag -a 1.0.0 -m "test message"')
        # Generate RELIC-INFO
        relic.release.get_info()
        shutil.rmtree('.git', onerror=onerror)
        # Read it back without git
        v = relic.release.get_info()
        assert v.short == '1.0.0'

    def test_read_relic_info_no_git(self):
        shutil.rmtree('.git', onerror=onerror)
        v = relic.release.get_info()
        assert '+' not in v.pep386
        assert v.short == '0.0.0'
        assert not v.commit



from setuptools import setup, find_packages
import relic.release

version = relic.release.get_info()
relic.release.write_template(version, 'relic/')

setup(
    name='relic',
    version=version.pep386,
    author='Joseph Hunkeler',
    author_email='jhunk@stsci.edu',
    description='Maintains version information for git projects',
    url='https://github.com/spacetelescope/relic',
    license='BSD',
    classifiers=[
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    packages=find_packages(),
    package_data={'': ['README.md', 'LICENSE.txt']}
)



import json
import os
from datetime import datetime
from . import ABBREV
from . import git


template = """# AUTOMATICALLY GENERATED BY 'RELIC':
# * DO NOT EDIT THIS MODULE MANUALLY.
# * DO NOT COMMIT THIS MODULE TO YOUR GIT REPOSITORY

__all__ = [
    '__version__',
    '__version_short__',
    '__version_long__',
    '__version_post__',
    '__version_commit__',
    '__version_date__',
    '__version_dirty__',
    '__build_date__',
    '__build_time__',
    '__build_status__'
]

__version__ = '{0}'
__version_short__ = '{1}'
__version_long__ = '{2}'
__version_post__ = '{3}'
__version_commit__ = '{4}'
__version_date__ = '{5}'
__version_dirty__ = {6}
__build_date__ = '{7}'
__build_time__ = '{8}'
__build_status__ = 'release' if not int(__version_post__) > 0 \\
    and not __version_dirty__ \\
    else 'development'
"""


def write_template(info, module_path, filename='version.py'):
    assert isinstance(info, git.GitVersion)
    path = os.path.join(module_path, filename)
    build_date = datetime.now().date()
    build_time = datetime.now().time()
    with open(path, 'w+') as f:
        output = template.format(
            info.pep386,
            info.short,
            info.long,
            info.post,
            info.commit,
            info.date,
            info.dirty,
            build_date,
            build_time
        )
        f.write(output)


def write_info(version):
    info = version

    if isinstance(version, git.GitVersion):
        info = version._asdict()

    with open('RELIC-INFO', 'w+') as f:
        f.write(json.dumps(info))
        f.write('\n')


def read_info():
    try:
        with open('RELIC-INFO', 'r') as f:
            data = json.loads(f.read())
            return git.GitVersion(**data)
    except (OSError, IOError):
        return None


def get_info(remove_pattern='release_'):
    relic_data = read_info()
    git_data = git.git_version_info(remove_pattern=remove_pattern)

    if not relic_data and not git_data:
        print('warning: no version data available!')
        return _fallback()

    elif git_data != relic_data and git_data:
        write_info(git_data)
        return git_data

    return relic_data


def _fallback():
    no_ver = '0.0.0'
    data = dict(
        pep386=no_ver,
        short=no_ver,
        long=no_ver,
        date='',
        dirty=True,
        commit='',
        post='-1',
    )
    return git.GitVersion(**data)



import re
from collections import namedtuple
from subprocess import Popen, PIPE
from . import PY3
from . import ABBREV


RE_GIT_DESC = re.compile('v?(.+?)-(\d+)-g(\w+)-?(.+)?')
GitVersion = namedtuple('GitVersion',
                        ['pep386', 'short', 'long', 'date', 'dirty', 'commit',
                         'post'])


def strip_dirty(tag):
    if tag.endswith('-dirty'):
        tag.replace('-dirty', '')
    return tag


def git(*commands):
    command = ['git', '--no-pager'] + [c for c in commands]
    proc = Popen(command, stdout=PIPE, stderr=PIPE, stdin=PIPE)

    if PY3:
        outs, errs = proc.communicate()
        if isinstance(outs, bytes):
            outs = outs.decode()
        if isinstance(errs, bytes):
            errs = errs.decode()
    else:
        outs, errs = proc.communicate()

    outs = outs.strip()
    errs = errs.strip()

    returncode = proc.wait()

    if returncode or errs:
        # Return code 128 implies we are trying to use git outside of a git
        # repository. This is a standard mode of operation for relic.
        if returncode == 128:
            return None

        print('{0} (exit: {1})'.format(errs, returncode))
        return None

    return outs


def git_describe(abbrev=ABBREV):
    return git('describe', '--always', '--long', '--tags', '--dirty',
               '--abbrev={0}'.format(abbrev))


def git_log_date(tag='HEAD'):
    tag = strip_dirty(tag)
    return git('log', '-1', '--format=%ai', tag)


def git_count(tag='HEAD'):
    tag = strip_dirty(tag)
    return git('rev-list', '--count', tag)


def git_version_info(remove_pattern=None):
    dirty = False
    commit = ''
    post = ''
    pep386 = ''
    version_short = ''
    version_long = git_describe()

    if not version_long:
        return None

    date = git_log_date(version_long)

    if isinstance(remove_pattern, str):
        if remove_pattern in version_long:
            version_long = version_long.replace(remove_pattern, '')
    elif isinstance(remove_pattern, list):
        for pattern in remove_pattern:
            version_long = version_long.replace(pattern, '')

    # Construct version data from repository
    match = RE_GIT_DESC.match(version_long)
    if match is not None:
        version_short, post, commit, dirty_check = match.groups()
        pep386 = version_short  # assume release version

        if dirty_check is not None:
            dirty = True

        if int(post):  # construct development version
            pep386 = '{}.dev{}+g{}'.format(version_short, post, commit)

    # No tag or not enough data to proceed
    else:
        if version_long.endswith('-dirty'):
            version_long = strip_dirty(version_long)
            dirty = True

        # Construct version data with what *might* be available
        commit = version_long
        version_long = version_short = '0.0.0'
        post = str(git_count()) or '-1'

        if int(post):  # construct development version
            pep386 = '{}.dev{}+g{}'.format(version_short, post, commit)

    data = dict(
        pep386=pep386,
        short=version_short,
        long=version_long,
        date=date,
        dirty=dirty,
        commit=commit,
        post=post,
    )

    return GitVersion(**data)
